/******************************************************************************
 * useful.c
 * The usual gang of helpers.
 *-----------------------------------------------------------------------------
 * Copyright (C) 2015 Osman Celimli
 * For conditions of distribution and use, see copyright notice in ctcc.c
 ******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>

#include "../types.h"

/******************************************************************************
 * Text Tables
 ******************************************************************************/
extern const char szERR_NoMainMem[];
extern const char szERR_Null[];
extern const char szERR_EarlyEnd[];

const char szERR_BadBinStr[] =	"ERROR- Invalid binary value : ";
const char szERR_BadNumStr[] =	"ERROR- Unable to parse numeric : ";

/******************************************************************************
 * int tealloc(void **pCurAddr, size_t *pCurLen, size_t reqLen, size_t elemSize)
 *  Attempt to reallocate the given block of memory whose base address is
 * stored at *pCurAddr if reqLen > *pCurLen. The pointer and block size will
 * automatically be updated if the reallocation occurs and is successful.
 *  The actual size of the block is calculated using elemSize, i.e. the
 * required length in bytes would be reqLen * elemSize.
 *
 * Returns nonzero on any failure.
 ******************************************************************************/
int tealloc(void **pCurAddr, size_t *pCurLen, size_t reqLen,
			size_t elemSize)
{
	void *pNewAddr;
	size_t newLen;

	if((NULL == pCurAddr) || (NULL == pCurLen))
	{
		fprintf(stderr, "%s", szERR_Null);
		return -1;
	}

	if(*pCurLen < reqLen)
	{
		newLen = ((*pCurLen << 1) > reqLen) ? (*pCurLen << 1) : reqLen;

		pNewAddr = realloc(*pCurAddr, (newLen * elemSize));
		if(NULL == pNewAddr)
		{
			fprintf(stderr, "%s", szERR_NoMainMem);
			return -1;
		}

		*pCurLen = newLen;
		*pCurAddr = pNewAddr;
	}
	return 0;
}




/******************************************************************************
 * int swapext(const char *pszSource, char **ppszDest, const char *pszExt, const char *pszTerm)
 *  Change the filename stored in pszSource to use the extension in pszExt,
 * storing the new filename in *pszDest. If no extension is present in the
 * original filename, a new one will be added.
 *  Termination characters for the backwards search of the original filename
 * can be specified using pszTerm.
 *
 * If any error occurs during, a nonzero value is returned.
 ******************************************************************************/
int swapext(const char *pszSource, char **ppszDest, const char *pszExt, const char *pszTerm)
{
	unsigned int uiLen;
	unsigned int uiX,uiY,uiZ;

	/**
	 * Make sure all of our argments are valid and then walk
	 * backwards through the string until we hit either
	 * a dot '.' indicating the presence of an extension
	 * or one of the supplied terminators.
	 */
	if((NULL == pszSource) || (NULL == ppszDest) || (NULL == pszExt))
	{
		fprintf(stderr, "%s", szERR_Null);
		return -1;
	}

	if((uiLen = strlen(pszSource)) < 1)
		return -1;

	uiX = uiLen - 1;
	uiY = uiX;
	while(uiX > 0)
	{
		if('.' == pszSource[uiX])
		{
			uiY = uiX;
			break;
		}

		if(NULL != strchr(pszTerm, pszSource[uiX]))
			break;
		uiX--;
	}

	uiZ = uiY + strlen(pszExt) + 1;
	*ppszDest = malloc(sizeof(char) * uiZ);
	if(NULL == *ppszDest)
	{
		fprintf(stderr, "%s", szERR_NoMainMem);
		return -1;
	}

	if(uiY > 0)
	{
		memcpy(*ppszDest, pszSource, (sizeof(char) * uiY));
	}
	strcpy((*ppszDest + uiY), pszExt);
	return 0;
}

/******************************************************************************
 * int strcut(char *pszTarget, char *pszRem, char cRep)
 *  Find groups of one of more members of pszRem in the string pszTarget and
 * replace them with a single character cRep.
 ******************************************************************************/
int strcut(char *pszTarget, char *pszRem, char cRep)
{
	char cCur;
	int iReplace;
	unsigned int uiL,uiR;

	if((NULL == pszTarget) || (NULL == pszRem))
	{
		fprintf(stderr, "%s", szERR_Null);
		return -1;
	}
	iReplace = -1;
	uiL = 0;
	uiR = 0;

	while('\0' != (cCur = pszTarget[uiR++]))
	{
		if(NULL == strchr(pszRem, cCur))
		{
			pszTarget[uiL++] = cCur;
			iReplace = -1;
		}
		else if(iReplace)
		{
			pszTarget[uiL++] = cRep;
			iReplace = 0;
		}
	}

	pszTarget[uiL] = '\0';
	if((uiL > 0) && (pszTarget[uiL - 1] == cRep))
	{
		pszTarget[uiL - 1] = '\0';
	}
	return 0;
}

/******************************************************************************
 * int strtoint(char *pszString, int *piTarget)
 *  Parse a string containing a value in either $hexadecimal, %binary, or
 * decimal. piTarget will not be overwritten if any errors occur during the
 * conversion process.
 ******************************************************************************/
int strtoint(char *pszString, int *piTarget)
{
	int iSuccess, iVal;
	unsigned int uiX;

	if((NULL == pszString) || (NULL == piTarget))
	{
		fprintf(stderr, "%s", szERR_Null);
		return -1;
	}

	if('\0' == pszString[0])
	{
		/* Empty
		 -----------------------------*/
		fprintf(stderr, "%s", szERR_EarlyEnd);
		return -1;
	}
	else if('$' == pszString[0])
	{
		/* $ Hexadecimal
		 -----------------------------*/
		iSuccess = sscanf((pszString + 1), "%x", &iVal);
	}
	else if('%' == pszString[0])
	{
		/* % Binary
		 -----------------------------*/
		iVal = 0;
		uiX = 1;

		while(('\0' != pszString[uiX]) && (uiX < 33))
		{
			iVal = iVal << 1;

			if(pszString[uiX] == '1') iVal |= 1;
			else if(pszString[uiX] != '0')
			{
				fprintf(stderr, "%s%s", szERR_BadBinStr, pszString);
				return -1;
			}
			uiX++;
		}
		if(uiX == 1)
		{
			fprintf(stderr, "%s%s", szERR_BadBinStr, pszString);
			return -1;
		}
		iSuccess = 1;
	}
	else
	{
		/* Decimal
		 -----------------------------*/
		iSuccess = sscanf(pszString, "%d", &iVal);
	}

	if((iSuccess != 0) && (iSuccess != EOF))
	{
		*piTarget = iVal;
	}
	else
	{
		fprintf(stderr, "%s%s", szERR_BadNumStr, pszString);
		return -1;
	}
	return 0;
}

/******************************************************************************
 * int bstrtoint(char *pszSrc, int *piVal)
 *  Extract an integer value from a string containing a series of zeroes and
 * ones. For example, "1111" would yield 15.
 *
 * Terminations will occur on a null terminator or '.'
 * Any errors encountered during parsing will yield a negative value.
 ******************************************************************************/
int bstrtoint(char *pszSrc, int *piVal)
{
	int iRot,iAcc;

	if(NULL == pszSrc)
	{
		fprintf(stderr, "%s", szERR_Null);
		return -1;
	}

	iAcc = 0;
	for(iRot = 0; iRot; (iRot = (iRot << 1)))
	{
		if(*pszSrc == '\0' || *pszSrc == '.') break;
		else if(*pszSrc == '1') iAcc = ((iAcc << 1) | 1);
		else if(*pszSrc == '0') iAcc = (iAcc << 1);
		else return -1;

		pszSrc++;
	}

	*piVal = iAcc;
	return 0;
}




/******************************************************************************
 * int parse16p16(int16p16_t *pstVal, char *pszSrc)
 *  Parse an input string and convert it to 16.16 format. Source can be in
 * hex ($xx.xx), decimal (xx.x), or binary (%xx.x)
 * If any error occurs during parsing, a nonzero value is returned.
 ******************************************************************************/
int parse16p16(int16p16_t *pstVal, char *pszSrc)
{
	unsigned int uiSrcLen;
	int iScan,iUpper,iLower;

	int16p16_t stVal;
	unsigned int uiX;

	/**
	 * Make sure both our source string and destination
	 * structure are valid, and that our source string
	 * has a length greater than zero.
	 */
	if((NULL == pszSrc) || (NULL == pstVal))
	{
		fprintf(stderr, "%s", szERR_Null);
		return -1;
	}

	iUpper = 0;
	iLower = 0;
	stVal.iWhole = 0;
	if((uiSrcLen = strlen(pszSrc)) < 1)
	{
		fprintf(stderr, "%s%s\n", szERR_BadNumStr, pszSrc);
		return -1;
	}

	/**
	 * $HEX.HEX
	 * %BIN.BIN
	 *  DEC.DEC
	 */
	if(pszSrc[0] == '$')
	{
		/* $HEX.HEX
		 -----------------------------*/
		iScan = sscanf(pszSrc + 1, "%x.%x", &iUpper, &iLower);
	}
	else if(pszSrc[0] == '%')
	{
		/* %BIN.BIN
		 -----------------------------*/
		for(uiX = 0; uiX < uiSrcLen; uiX++)
		{
			if(pszSrc[uiX] == '.') break;
		}
		if(bstrtoint(pszSrc + 1, &iUpper) || bstrtoint(pszSrc + uiX + 1, &iLower))
		{
			fprintf(stderr, "%s%s\n", szERR_BadNumStr, pszSrc);
			return -1;
		}

		iScan = 2;
	}
	else
	{
		/* DEC.DEC
		 -----------------------------*/
		iScan = sscanf(pszSrc, "%d.%d", &iUpper, &iLower);
		if(pszSrc[0] == '-')
		{
			if(!((iScan == 1) || ((iScan == 2) && (iLower == 0))))
			{
				iUpper -= 1;
			}
		}
	}

	/**
	 * Clamp our value to a 16.16 swing, ensuring we actually
	 * parsed something and that our decimal (if present) was
	 * actually positive.
	 */
	if((iScan == 0) || (iLower < 0))
	{
		fprintf(stderr, "%s%s\n", szERR_BadNumStr, pszSrc);
		return -1;
	}

	iUpper = (iUpper > 65535)
		? 65535
		: ((iUpper < -32768) ? -32768 : iUpper);
	iLower = (iUpper < 0)
		? (65536 - iLower)
		: iLower;

	pstVal->usPair.usHi = iUpper;
	pstVal->usPair.usLo = iLower;
	return 0;
}

/******************************************************************************
 * int parse16p8(int16p16_t *pstVal, char *pszSrc)
 *  Parse an input string and convert it to 16.8 format, calling parse16p16
 * under the hood and then just clamping its output.
 ******************************************************************************/
int parse16p8(int16p16_t *pstVal, char *pszSrc)
{
	int16p16_t stVal;

	if(parse16p16(&stVal, pszSrc)) return -1;

	stVal.sPair.sLo &= 0xFF;
	pstVal->iWhole = stVal.iWhole;
	return 0;
}
