/******************************************************************************
 * playtone.cpp
 * CoreTone Auditioning Suite for Windows.
 *-----------------------------------------------------------------------------
 * Version 1.1cz, December 2nd, 2015
 * Copyright (C) 2015 Osman Celimli
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 ******************************************************************************/
#include <Windows.h>
#include <Commctrl.h>
#include <stdint.h>
#include <tchar.h>

#include "../types.h"
#include "../WinTone/wintone.hpp"
#include "../CoreTone/channel.h"
#include "../CoreTone/music.h"

#include "toneman.hpp"
#include "menuman.hpp"
#include "statman.hpp"
#include "chanman.hpp"
#include "resource.h"

/******************************************************************************
 * !!!!----              SYSTEM STATE and RENDER INFO                ----!!!!
 ******************************************************************************/
ToneMan *pToneMan = NULL;
MenuMan *pMenuMan = NULL;
StatMan *pStatMan = NULL;
ChanMan *pChanMan = NULL;

UINT uiCoreChannels;
UINT uiRenderFreq,uiDecodeRate;
UINT uiSamples,uiSampleLen;

CoreChannel_t *pCoreChannels = NULL;
CorePatch_t *pCorePatches = NULL;
CoreTrack_t *pCoreTracks = NULL;

/******************************************************************************
 * !!!!----                   WINDOWS GUI GIBBUMS                    ----!!!!
 ******************************************************************************/
#define DRAW_TIMER_ID			101
#define DRAW_TIMER_MS			50

/* LoadString Buffer
 *-----------------------------------*/
#define STRING_BUFFER_ENTRIES	512
#define STRING_BUFFER_LEN		(sizeof(szBuffer) / sizeof(szBuffer[0]))

TCHAR szBuffer[STRING_BUFFER_ENTRIES];


/* Main Window and Friends
 *-----------------------------------*/
const TCHAR szWinClass[] = _T("ToneClass");
const TCHAR szWinName[] = _T("PlayTone");
WNDCLASSEX wcWinMain;

HWND hMainWnd = NULL;
UINT uiMainWidth,uiMainHeight;
UINT_PTR pDrawTimer = NULL;


/******************************************************************************
 * !!!!----                   WINMAIN and WINPROC                    ----!!!!
 ******************************************************************************/
/* static LRESULT CALLBACK WinProc(HWND hWnd, UINT uMsg,
 *                                 WPARAM wParam, LPARAM lParam)
 * Window process...
 *----------------------------------------------------------------------------*/
static LRESULT CALLBACK WinProc(HWND hWnd, UINT uMsg,
                                WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT psWin;
	HDC hDC;

	switch(uMsg)
	{
		/**
		 * ---- WM_DESTROY ----
		 * The window is being destroyed. Note that we'll only receive
		 * this message once all of this window's children have already
		 * been synchronously destroyed.
		 *
		 * This means that once we post the quit messaage to WinMain's
		 * message loop, it is guaranteed that no windows are active.
		 */
		case WM_DESTROY:
			if(NULL != pDrawTimer)
				KillTimer(hWnd, DRAW_TIMER_ID);

			PostQuitMessage(0);
			break;

		/**
		 * ---- WM_CLOSE ----
		 * The application is being terminated, this is the message
		 * received when the user clicks the little "X".
		 */
		case WM_CLOSE:
			DestroyWindow(hWnd);
			break;

		/**
		 * ---- WM_ERASEBKGND ----
		 * The window background must be erased (and the window
		 * has possibly been resized). We should return a nonzero
		 * value as long as the background is erased properly.
		 */
		case WM_ERASEBKGND:
			break;

		/**
		 * ---- WM_TIMER ----
		 * The timer event associated with this window has expired,
		 * we're using this to time draw events once the driver is running.
		 */
		case WM_TIMER:
			if(pToneMan->mbDriverRunning)
			{
				pStatMan->InvalProc(hWnd);
				pChanMan->InvalProc(hWnd);

				UpdateWindow(hWnd);
			}
			break;

		/**
		 * ---- WM_PAINT ----
		 * Time to redraw!
		 */
		case WM_PAINT:
			/**
			 * Before we perform any painting, we need to mark the start
			 * of our draw session and ensure the display device context
			 * can actually be established.
			 */
			hDC = BeginPaint(hWnd, &psWin);
			if(NULL != hDC)
			{
				pStatMan->PaintProc(hWnd, hDC, psWin);
				pChanMan->PaintProc(hWnd, hDC, psWin);

				EndPaint(hWnd, &psWin);
			}
			break;

		/**
		 * ---- WM_ENTERMENULOOP and WM_EXITMENULOOP ----
		 * A menu modal loop has been entered or exited (a menu
		 * has been opened or closed respectively).
		 */
		case WM_ENTERMENULOOP:
			break;
		case WM_EXITMENULOOP:
			break;

		/**
		 * ---- WM_COMMAND ----
		 * Any of the menu options (IDM_x) have been selected or
		 * an accelerator key combination has been pressed.
		 */
		case WM_COMMAND:
			return (pMenuMan->CmdProc(hWnd, uMsg, wParam, lParam)
				| pStatMan->CmdProc(hWnd, uMsg, wParam, lParam));

		/**
		 * ---- WM_HSCROLL ----
		 * An interaction with a horizontal scrollbar or trackbar
		 * has been performed.
		 * wParam indicates the type of interaction while lParam
		 * is the handle of who was interacted with.
		 */
		case WM_HSCROLL:
			return pStatMan->TrackProc(hWnd, uMsg, wParam, lParam);

		/**
		 * How did we get here?
		 * Windows, figure it out for us.
		 */
		default:
			break;
	}

	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}




/* int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
 *                    LPSTR lpCmdLine, int nShowCmd)
 * WinMain...
 * It's Main for Wins... yeah.
 *----------------------------------------------------------------------------*/
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
				   LPSTR lpCmdLine, int nShowCmd)
{
	HICON hIcon32,hIcon16;
	HICON hCursor;
	HACCEL hAccels;
	MSG MainMsg;

	INT iMisMatch;
	INT iStatus;


	/**
	 * Get the properties of our loaded WinTone + CoreTone DLLs, ensuring
	 * both of them are paired appropriately before starting the application.
	 *
	 * We'll also fetch the internal CoreTone structure addresses at this time
	 * so the UI can display verbose driver activity.
	 */
	iMisMatch = wt_getInfo(&uiCoreChannels, &uiRenderFreq, &uiDecodeRate,
                           &uiSamples, &uiSampleLen);
	if(iMisMatch)
	{
		iStatus = LoadString(hInstance, IDS_CORE_MISMATCH,
			szBuffer, STRING_BUFFER_LEN);
		if(iStatus > 0)
		{
			MessageBox(NULL, szBuffer, NULL, (MB_OK | MB_ICONEXCLAMATION));
		}

		return 0;
	}

	wt_getState((void**)&pCoreChannels, (void**)&pCorePatches,
		(void**)&pCoreTracks, NULL, NULL, NULL, NULL);


	/**
	 * Load up our icons and cursors, these will be automatically freed
	 * when this program terminates, but according to the MSDN it
	 * is good manners to release them using DestroyIcon / DestroyCursor.
	 *
	 * Note that since we're using the default system cursor,
	 * we shouldn't destroy it.
	 */
	hIcon32 = (HICON)LoadImage(
		hInstance,
		MAKEINTRESOURCE(IDI_CORETONE),
		IMAGE_ICON,
		32, 32,
		LR_DEFAULTCOLOR);

	hIcon16 = (HICON)LoadImage(
		hInstance,
		MAKEINTRESOURCE(IDI_CORETONE),
		IMAGE_ICON,
		16, 16,
		LR_DEFAULTCOLOR);

	hCursor = LoadCursor(NULL, IDC_ARROW);


	/**
	 * ---- Window Class Setup ----
	 * Every window has a window class which defines basic appearance
	 * and system behavior within it such as which mouse cursor to use.
	 *
	 * These all must be registered using RegisterClassEx prior to
	 * creating any actual windows.
	 */
	wcWinMain.cbSize = sizeof(WNDCLASSEX);
	wcWinMain.style = 0;
	wcWinMain.lpfnWndProc = WinProc;
	wcWinMain.cbClsExtra = 0;
	wcWinMain.cbWndExtra = 0;
	wcWinMain.hInstance = hInstance;
	wcWinMain.hIcon = hIcon32;
	wcWinMain.hCursor = hCursor;
	wcWinMain.hbrBackground = (HBRUSH)(COLOR_WINDOW + 0);
	wcWinMain.lpszMenuName = MAKEINTRESOURCE(IDR_MENU_MAIN);
	wcWinMain.lpszClassName = szWinClass;
	wcWinMain.hIconSm = hIcon16;

	if(!RegisterClassEx(&wcWinMain))
	{
		iStatus = LoadString(hInstance, IDS_ERR_WINCLASS,
			szBuffer, STRING_BUFFER_LEN);
		if(iStatus > 0)
		{
			MessageBox(NULL, szBuffer, NULL, (MB_OK | MB_ICONEXCLAMATION));
		}

		return 0;
	}


	/**
	 * ---- Main Window Creation ----
	 * Now we can actually initialize our main window by supplying
	 * CreateWindowEx with a metric boatload of arguments.
	 *
	 * All our text, buttons, and pretty graphs will be sub windows
	 * within this main window.
	 */
	InitCommonControls();
	uiMainWidth = 512;
	uiMainHeight = 48
		+ WSTAT_HEIGHT
		+ (CSTAT_HEIGHT_BASE + (CSTAT_HEIGHT_PER_CHANNEL * (uiCoreChannels + 1)));

	hMainWnd = CreateWindowEx(
		(0),									/* dwExStyle */
		szWinClass,								/* lpClassName */
		szWinName,								/* lpWindowName */
		(WS_SYSMENU | WS_CAPTION | WS_VISIBLE),	/* dwStyle */
		CW_USEDEFAULT,							/* x */
		CW_USEDEFAULT,							/* y */
		uiMainWidth,							/* nWidth */
		uiMainHeight,							/* nHeight */
		NULL,									/* hWndParent */
		NULL,									/* hMenu, NULL = Use CLASS Menu */
		hInstance,								/* hInstance */
		NULL);									/* lpParam */

	if(NULL == hMainWnd)
	{
		iStatus = LoadString(hInstance, IDS_ERR_WINCREATE,
			szBuffer, STRING_BUFFER_LEN);
		if(iStatus > 0)
		{
			MessageBox(NULL, szBuffer, NULL, (MB_OK | MB_ICONEXCLAMATION));
		}

		return 0;
	}


	/**
	 * ---- Message Loop ----
	 * Initialize the rest of our components which depend upon the
	 * main window, then sit in our dispatch loop until the user quits
	 * or something self-destructs.
	 */
	pToneMan = new ToneMan(hMainWnd);
	pMenuMan = new MenuMan(hMainWnd, pToneMan);

	pStatMan = new StatMan(hMainWnd, pToneMan);
	pStatMan->muiCoreChannels = uiCoreChannels;
	pStatMan->muiDecodeRate = uiDecodeRate;
	pStatMan->muiRenderFreq = uiRenderFreq;
	pStatMan->muiSamples = uiSamples;
	pStatMan->muiSampleLen = uiSampleLen;

	pChanMan = new ChanMan(hMainWnd, pCoreChannels, pCorePatches, uiCoreChannels);

	pDrawTimer = SetTimer(hMainWnd, DRAW_TIMER_ID, DRAW_TIMER_MS, NULL);

	hAccels = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDR_ACCELERATOR));
	while(GetMessage(&MainMsg, NULL, 0, 0) > 0)
	{
		if(!TranslateAccelerator(hMainWnd, hAccels, &MainMsg))
		{
			TranslateMessage(&MainMsg);
			DispatchMessage(&MainMsg);
		}
	}


	/**
	 * ---- Exit ----
	 * Clean up and we're outta here!
	 */
	if(NULL != hIcon32)
		DestroyIcon(hIcon32);
	if(NULL != hIcon16)
		DestroyIcon(hIcon16);

	delete pMenuMan;
	pMenuMan = NULL;
	delete pToneMan;
	pToneMan = NULL;
	delete pStatMan;
	pStatMan = NULL;
	delete pChanMan;
	pChanMan = NULL;

	return MainMsg.wParam;
}
