/******************************************************************************
 * toneman.cpp
 * Secondary wrapper around WinTone + CoreTone to manage resources.
 *-----------------------------------------------------------------------------
 * Copyright (C) 2015 Osman Celimli
 * For conditions of distribution and use, see copyright notice in playtone.cpp
 ******************************************************************************/
#include <Windows.h>
#include <stdint.h>
#include <stdio.h>

#include "../WinTone/wintone.hpp"
#include "toneman.hpp"

/******************************************************************************
 * !!!!----               CONSTRUCTOR AND DESTRUCTOR                 ----!!!!
 ******************************************************************************/
ToneMan::ToneMan(HWND hParent)
{
	mhParent = hParent;
	mbDriverRunning = FALSE;
	mbLogging = FALSE;

	mbSamplesReady = FALSE;
	mbInstrumentsReady = FALSE;
	mpSamples = NULL;
	mpInstruments = NULL;

	mbMusicReady = FALSE;
	mpMusic = NULL;

	mbSfxReady = FALSE;
	mpSfx = NULL;
	miLastPriority = 0;
}

ToneMan::~ToneMan(void)
{
	StopDriver();

	if(mbSfxReady && (NULL != mpSfx))
	{
		free(mpSfx);
	}
	mbSfxReady = FALSE;
	mpSfx = NULL;

	if(mbMusicReady && (NULL != mpMusic))
	{
		free(mpMusic);
	}
	mbMusicReady = FALSE;
	mpMusic = NULL;

	if(mbInstrumentsReady && (NULL != mpInstruments))
	{
		free(mpMusic);
	}
	mbInstrumentsReady = FALSE;
	mpInstruments = NULL;

	if(mbSamplesReady && (NULL != mpSamples))
	{
		free(mpMusic);
	}
	mbSamplesReady = FALSE;
	mpSamples = NULL;
}




/******************************************************************************
 * !!!!----               DRIVER STARTUP and SHUTDOWN                ----!!!!
 ******************************************************************************/
/* HRESULT ToneMan::StartDriver(void)
 *  Attempt to start the driver if and only if the sample and instrument
 * packages have been loaded and the driver is not already running.
 *----------------------------------------------------------------------------*/
HRESULT ToneMan::StartDriver(void)
{
	HRESULT hRes = E_FAIL;

	if(!mbDriverRunning && mbSamplesReady && mbInstrumentsReady)
	{
		hRes = wt_init(mhParent, mpSamples, mpInstruments);
		mbDriverRunning = FAILED(hRes)
			? FALSE
			: TRUE;
	}

	return hRes;
}

/* HRESULT ToneMan::StopDriver(void)
 *  Close the driver.
 *----------------------------------------------------------------------------*/
HRESULT ToneMan::StopDriver(void)
{
	HRESULT hRes = E_FAIL;

	if(mbDriverRunning)
	{
		wt_stopAll();
		wt_close();
		mbDriverRunning = FALSE;
		mbLogging = FALSE;

		hRes = S_OK;
	}

	return hRes;
}




/******************************************************************************
 * !!!!----        SAMPLE, INSTRUMENT, MUSIC, and SFX LOADING        ----!!!!
 ******************************************************************************/
/* HRESULT ToneMan::LoadFile(LPSTR pszFile, BYTE **ppDest)
 *  General use binary file loader, will attempt to load the file whose name
 * is in pszFile to a newly malloc'd block of memory whose base address will
 * be written to *ppDest.
 *----------------------------------------------------------------------------*/
HRESULT ToneMan::LoadFile(LPSTR pszFile, BYTE **ppDest)
{
	HRESULT hRes = E_ACCESSDENIED;

	FILE *pFile = NULL;
	BYTE *pData = NULL;
	UINT uiLen,uiReadLen;


	if(NULL != (pFile = fopen(pszFile, "rb")))
	{
		fseek(pFile, 0, SEEK_END);
		uiLen = ftell(pFile);
		rewind(pFile);

		if(NULL == (pData = (BYTE*)malloc(uiLen)))
		{
			hRes = E_OUTOFMEMORY;
		}
		else
		{
			uiReadLen = fread(pData, sizeof(char), uiLen, pFile);
			fclose(pFile);

			if(uiReadLen == uiLen)
			{
				*ppDest = pData;
				hRes = S_OK;
			}
			else
			{
				free(pData);
				hRes = E_UNEXPECTED;
			}
		}
	}

	return hRes;
}

/* HRESULT ToneMan::LoadSamples(LPSTR pszFile)
 *  Attempt to load the sample package whose filename is in pszFile, note that
 * this will automatically close the driver if it is currently running.
 *----------------------------------------------------------------------------*/
HRESULT ToneMan::LoadSamples(LPSTR pszFile)
{
	HRESULT hRes = E_FAIL;


	if(mbDriverRunning)
	{
		StopDriver();
	}

	if(mbSamplesReady)
	{
		free(mpSamples);
		mpSamples = NULL;
		mbSamplesReady = FALSE;
	}

	hRes = LoadFile(pszFile, &mpSamples);
	mbSamplesReady = FAILED(hRes)
		? FALSE
		: TRUE;

	return hRes;
}

/* HRESULT ToneMan::LoadInstruments(LPSTR pszFile)
 *  Attempt to load the instrument package whose filename is in pszFile, note
 * that this will automatically close the driver if it is currently running.
 *----------------------------------------------------------------------------*/
HRESULT ToneMan::LoadInstruments(LPSTR pszFile)
{
	HRESULT hRes = E_FAIL;


	if(mbDriverRunning)
	{
		StopDriver();
	}

	if(mbInstrumentsReady)
	{
		free(mpInstruments);
		mpInstruments = NULL;
		mbInstrumentsReady = FALSE;
	}

	hRes = LoadFile(pszFile, &mpInstruments);
	mbInstrumentsReady = FAILED(hRes)
		? FALSE
		: TRUE;

	return hRes;
}

/* HRESULT ToneMan::LoadMusic(LPSTR pszFile)
 *  Attempt to load the music track whose filename is in pszFile, this will
 * automatically cease playback of any currently active music.
 *----------------------------------------------------------------------------*/
HRESULT ToneMan::LoadMusic(LPSTR pszFile)
{
	HRESULT hRes = E_FAIL;


	if(mbMusicReady)
	{
		StopMusic();

		free(mpMusic);
		mpMusic = NULL;
		mbMusicReady = FALSE;
	}

	hRes = LoadFile(pszFile, &mpMusic);
	mbMusicReady = FAILED(hRes)
		? FALSE
		: TRUE;

	return hRes;
}

/* HRESULT ToneMan::LoadSfx(LPSTR pszFile)
 *  Attempt to load the sound effect whose filename is in pszFile, this will
 * automatically cease playback of any currently active sound effect.
 *----------------------------------------------------------------------------*/
HRESULT ToneMan::LoadSfx(LPSTR pszFile)
{
	HRESULT hRes = E_FAIL;


	/**
	 * All sound effects dispatched from the MENU have a default priority of
	 * 127 and full volume center panning. However, using the sound effect
	 * manager to dispatch more instances of the currently loaded effect
	 * will overwrite the cached priority from the menu.
	 *
	 * So before we load, we'll stop both the sound effect with the last
	 * logged priority and the menu priority of 127.
	 * This isn't an issue for game software, which is expected to have
	 * all sound effects loaded up front.
	 */
	if(mbSfxReady)
	{
		StopSfx();
		miLastPriority = 127;
		StopSfx();

		free(mpSfx);
		mpSfx = NULL;
		mbSfxReady = FALSE;
	}

	hRes = LoadFile(pszFile, &mpSfx);
	mbSfxReady = FAILED(hRes)
		? FALSE
		: TRUE;

	return hRes;
}




/******************************************************************************
 * !!!!----                      *.WAV LOGGING                       ----!!!!
 ******************************************************************************/
HRESULT ToneMan::StartLog(LPSTR pszFile)
{
	HRESULT hRes = E_FAIL;

	if(mbDriverRunning && !mbLogging)
	{
		if(0 == wt_logOutput(pszFile))
		{
			mbLogging = TRUE;
			hRes = S_OK;
		}
	}

	return hRes;
}

HRESULT ToneMan::StopLog(void)
{
	HRESULT hRes = E_FAIL;

	if(mbDriverRunning && mbLogging)
	{
		wt_closeLog();

		mbLogging = FALSE;
		hRes = S_OK;
	}

	return hRes;
}




/******************************************************************************
 * !!!!----                     PLAYBACK CONTROL                     ----!!!!
 ******************************************************************************/
void ToneMan::Pause(void)
{
	if(mbDriverRunning)
	{
		wt_pause();
	}
}

void ToneMan::Resume(void)
{
	if(mbDriverRunning)
	{
		wt_resume();
	}
}

void ToneMan::StopAll(void)
{
	if(mbDriverRunning)
	{
		wt_stopAll();
	}
}

void ToneMan::PlayMusic(void)
{
	if(mbDriverRunning && mbMusicReady)
	{
		wt_playMusic(mpMusic);
	}
}

void ToneMan::StopMusic(void)
{
	if(mbDriverRunning && mbMusicReady)
	{
		wt_stopMusic();
	}
}

void ToneMan::AttenMusic(BYTE cVol)
{
	if(mbDriverRunning && mbMusicReady)
	{
		wt_attenMusic(cVol);
	}
}

void ToneMan::PlaySfx(INT iPriority, INT iVol_Left, INT iVol_Right)
{
	if(mbDriverRunning && mbSfxReady && (0 != iPriority))
	{
		miLastPriority = iPriority;
		wt_playSFX(mpSfx, miLastPriority, iVol_Left, iVol_Right);
	}
}

void ToneMan::StopSfx(void)
{
	if(0 != miLastPriority)
	{
		wt_stopSFX(miLastPriority);
	}
}
